// Copyright (C) 1998 Logi Ragnarsson

package is.logi.crypto.keys;
import is.logi.crypto.*;
import is.logi.crypto.hash.*;

import java.math.BigInteger;
import java.io.*;

/**
 * This object holds one Diffie-Hellman key. It is not used for encryption
 * or signatures, but only to initialize Difie-Hellman key-exchange objects.
 * <p>
 * CDS for a Diffie-Hellman key is <code>DHKey(a,g,m,pub)</code>
 * for a public key or <code>DHKey(a,g,m,pri)</code> for a private
 * key. In both cases <code>a</code>,<code>g</code> and <code>m</code> are
 * hexadecimal numbers.
 *
 * @see is.logi.crypto.protocols.DHKeyExNoninter
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class DHKey extends K implements Key {
    
    /** Actual key value. */
    private BigInteger a;
    
    /** Generator */
    private BigInteger g;
    
    /** Modulus */
    private BigInteger m;
    
    /** Private-key flag. */
    private boolean pri;
    
    /**
     * If <code>pri</code>, then this is the public key
     * or <code>null</code>.
     */
    private BigInteger pubKey;
    
    /**
     * Create a new Diffie-Hellman key. An object is created for <code>a</code>
     * in the group modulo <code>m</code> with generator <code>g</code>. It is
     * a private key iff </code>pri<code> is <code>true</code>.
     */
    public DHKey(BigInteger a, BigInteger g, BigInteger m, boolean pri){
        this.a=a;
        this.g=g;
        this.m=m;
        this.pri=pri;
    }
    
    /**
     * If "DHKey( key )" is a valid CDS for a DHKey, then
     * DHKey.parseCDS(key) will return the described DHKey object.
     * <p>
     * A valid CDS can be created by calling the DHKey.toString() method.
     *
     * @exception InvalidCDSException if the CDS is malformed.
     *
     * @see is.logi.crypto.Crypto#fromString(String)
     */
    public static DHKey parseCDS(String key) throws InvalidCDSException{
        DHKey ret=null;
        try {
            StreamTokenizer st=new StreamTokenizer(new StringReader(key));
            st.ordinaryChars('0','9');
            st.wordChars('0','9');
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as first argument to DHKey()");
            BigInteger a = new BigInteger(st.sval,16);
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+a.toString(16));
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as second argument to DHKey()");
            BigInteger g = new BigInteger(st.sval,16);
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+g.toString(16));
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as third argument to DHKey()");
            BigInteger m = new BigInteger(st.sval,16);
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+m.toString(16));
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("\"pub\" or \"pri\" expected as second argument to DHKey()");
            String pubpri=st.sval;
            if(pubpri.equals("pri"))
                ret=new DHKey(a,g,m,true);
            else if(pubpri.equals("pub"))
                ret=new DHKey(a,g,m,false);
            else
                throw new InvalidCDSException("\"pub\" or \"pri\" expected as second argument to DHKey()");
            if(st.nextToken()!=StreamTokenizer.TT_EOF){
                throw new InvalidCDSException("no more parameters expected in DHKey() after "+pubpri);
            }
        } catch (IOException e){
            // StringReader doesn't really throw exceptions.
        }
        return ret;
    }
    
    /**
     * Return a Diffie-Hellman modulus.
     * <p>
     * Return the largest prime <code>p</code> &lt; 2^<code>n</code> such
     * that (<code>p</code>-1)/2 is prime. This takes a long time unless
     * <code>n</code> is one of 256, 512, 1024 or 2048.
     * <p> 
     * The reason for the second constraint on <code>p</code> is to make
     * the Discrete-Logarithm problem harder in the group Zp. In
     * particular it thwarts the Pholig-Hellman algorithm.
     */
    public static BigInteger getModulus(int n){
        if(n<256)
            n=256;
        else
            n=((n+7)/8)*8; // round up to nearest byte.
        
        if (n==256) 
            return TWO.pow(n).subtract(BigInteger.valueOf(36113));
        if (n==512)
            return TWO.pow(n).subtract(BigInteger.valueOf(38117));
        if (n==1024)
            return TWO.pow(n).subtract(BigInteger.valueOf(1093339));
        if (n==2048)
            return TWO.pow(n).subtract(BigInteger.valueOf(1942289));

        // The user asked for a non-standard modulus-size. We will
        // comply, but he probably won't have the patience to wait.

        BigInteger m=TWO.pow(n).subtract(ONE);
        //int sub=1;
        BigInteger k=m.divide(TWO);
        while(!m.isProbablePrime(2) || // First check very quickly
              !k.isProbablePrime(2) ||
              !m.isProbablePrime(4) ||
              !k.isProbablePrime(4) ||
              !m.isProbablePrime(8) ||
              !k.isProbablePrime(8) ||
              !m.isProbablePrime(primeCertainty-14)|| // Then check very well
              !k.isProbablePrime(primeCertainty-14)){
            m=m.subtract(FOUR);
            //sub+=4;
            k=m.divide(TWO);
        }
        return m;
    }
    
    /**
     * Return a Generator for a modulus group.
     * <p>
     * Returns the smallest number <code>g</code> > 1 with
     * gcd(<code>g</code>,<code>m-1</code>)=1.
     */
    public static BigInteger getGenerator(BigInteger m){
        BigInteger g=TWO;
        BigInteger n=m.subtract(ONE);
        while(!n.gcd(g).equals(ONE))
            g=g.add(ONE);
        return g;
    }
    
    /**
     * Create a random private DHKey with an <code>n</code>
     * bit modulus.
     * <p>
     * Pre-calcualted modula exist for bit lengths 256, 512, 1024 and 2048.
     * Using these values saves a lot of time and does not weaken the keys.
     */
    public DHKey(int n){
        n = (n<256) ? 256 : (((n+7)/8)*8); // round up to nearest byte.
        
        m=getModulus(n);
        g=getGenerator(m);
        do{
            a=new BigInteger(n,random);
        } while (a.compareTo(m)>=0);
        
        pri=true;
    }
    
    /**
     * Create a pair of public/private keys in a group with an
     * <code>n</code> bit modulo.
     * <p>
     * Pre-calcualted modula exist for bit lengths 256, 512, 1024 and
     * 2048. Using these values saves a lot of time and does not weaken
     * the keys.
     */
    public static KeyPair createKeys(int n){
        DHKey pri = new DHKey(n);
        return new KeyPair(pri.getPublic(),
                           pri);
    }
    
    /**
     * Return the "size" of the key. This is a (fairly inaccurate) measure
     * of how difficult it is to break and is heavily dependant on the
     * algorithm used.
     */
    public int getSize(){
        return m.bitLength();
    }
    
    /**
     * The name of the algorithm is "Diffie-Hellman".
     */
    public String getAlgorithm(){
        return "Diffie-Hellman";
    }
    
    /**
     * Returns true iff this is a private key.
     */
    public boolean isPrivate(){
        return pri;
    }
    
    /**
     * Return the public key from the pair.
     */
    public DHKey getPublic(){
        if(!pri)
            return this;
        if(pubKey==null)
            pubKey = g.modPow(a,m);
        return new DHKey(pubKey,g,m,false);
    }
    
    /**
     * Return the BigInteger representing this key.
     */
    public BigInteger getKey(){
        return a;
    }
    
    /**
     * Return the modulus for this key.
     */
    public BigInteger getM(){
        return m;
    }
    
    /**
     * Return the generator for this key.
     */
    public BigInteger getG(){
        return g;
    }
    
    /** 
     * Calculate the fingerprint for this key or the other in the pair.
     */
    protected Fingerprint calcFingerprint(boolean other){
        // The key-pair is uniquely defined by the public key, since the
        // private key is "merely" the dicrete log of the public key.
        HashState fs=new SHA1State();
        fs.update(getPublic().getKey().toByteArray());
        fs.update(getPublic().getM().toByteArray());
        if(other==pri)
            fs.update("pub");
        else
            fs.update("pri");
        return fs.calculate();
    }
    
    /**
     * Return true iff the two keys are equivalent.
     */
    public final boolean equals(Object o){
        if (o==null)
            return false;
        if (o.getClass() != this.getClass())
            return false;
        DHKey key=(DHKey)o;
        return (key.getKey().equals(getKey()) &&
                key.getM().equals(getM())     &&
                key.getG().equals(getG())
               );
    }
    
    /**
     * Check if a key mathces this. This is true if this and key are a matched
     * pair of public/private.
     */
    public boolean matches(Key key){
        if(key instanceof DHKey)
            return getPublic().equals(((DHKey)key).getPublic());
        else return false;
    }
    
    /**
     * Return a CDS for this key.
     */
    public String toString(){
        StringBuffer sb=new StringBuffer();
        sb.append("DHKey(");
        sb.append(a.toString(16));
        sb.append(',');
        sb.append(g.toString(16));
        sb.append(',');
        sb.append(m.toString(16));
        sb.append(pri ? ",pri)" : ",pub)");
        return sb.toString();
    }
    
}
